import tensorflow as tf
import numpy as np

from base_model import BaseModel

from test import randomly_construct_tf_model

tf_dtype_to_np_dtype = {
    tf.float16: np.float16,
    tf.float32: np.float32,
    tf.float64: np.float64,
    tf.int8: np.int8,
    tf.int16: np.int16,
    tf.int32: np.int32,
    tf.int64: np.int64,
    tf.uint8: np.uint8,
    tf.uint16: np.uint16,
    tf.uint32: np.uint32,
    tf.uint64: np.uint64,
}

class TensorflowSavedModel(BaseModel):
    """Tensorflow Saved Model object.

    Args:
        BaseModel (BaseModel): The base model object.
    """
    def __init__(self, config: dict, model, input_shapes: dict, output_shapes: dict):
        """ Initialize a tensorflow saved model object.

        Args:
            config: A dictionary containing the model configuration.
            model: The model object.
            input_shapes: A dictionary containing the input shapes and types of the model.
            output_shapes: A dictionary containing the output shapes and types of the model.
        """
        
        super().__init__(config, model, "Tensorflow-SavedModel", input_shapes, output_shapes)
    
    @classmethod
    def init(cls, config: dict, model) -> 'TensorflowSavedModel':
        """ A constructor-like function to init a tensorflow saved model object.

        Args:
            config: A dictionary containing the model configuration.
            model: The model object.
            input_shapes_and_types: A dictionary containing the input shapes and types of the model.
            output_shapes_and_types: A dictionary containing the output shapes and types of the model.
        """            
        input_specs = []
        output_specs = []
        
        def format_shape(shape_list):
            for i in range(len(shape_list)):
                if shape_list[i] == -1:
                    shape_list[i] = None
            return shape_list
        
        for value in model.get_input_details():
            input_specs.append({
                'name': value['name'],
                'index': value['index'],
                'shape': format_shape(list(value['shape_signature'])),
                'dtype': value['dtype'],
                'quantization_specs':{
                    'quantization': value['quantization'],
                    'quantization_parameters': value['quantization_parameters'],
                    'sparsity_parameters': value['sparsity_parameters'],
                }
            })
        
        for value in model.get_output_details():
            output_specs.append({
                'name': value['name'],
                'index': value['index'],
                'shape': format_shape(list(value['shape_signature'])),
                'dtype': value['dtype'],
                'quantization_specs':{
                    'quantization': value['quantization'],
                    'quantization_parameters': value['quantization_parameters'],
                    'sparsity_parameters': value['sparsity_parameters'],
                }
            })
            
        return cls(config, model, input_specs, output_specs)
    
if __name__ == "__main__":
    model, input_shapes, output_shapes = randomly_construct_tf_model()
    converter = tf.lite.TFLiteConverter.from_keras_model(model)
    tflite_model = converter.convert()

    # Save the model.
    with open('./model.tflite', 'wb') as f:
        f.write(tflite_model)

    # Load the TFLite model in TFLite Interpreter
    interpreter = tf.lite.Interpreter('./model.tflite')

    model = TensorflowSavedModel.init({}, interpreter)
    print(model.input_shapes)
    print(model.output_shapes)


import tensorflow as tf
import numpy as np

class BaseModel:
    """ Base model object. 
    
        Each model object should contain the following attributes:
        
        config: A dictionary containing the model configuration.
        model: The model object.
        model_type: The type of the model object.
        Input shapes and types: A dictionary containing the input shapes and types of the model.
        Output shapes and types: A dictionary containing the output shapes and types of the model.
    """
    
    def __init__(self, config: dict, model, model_type: str, input_shapes: dict, output_shapes: dict):
        """ Initialize a base model object.
        
        Args:
            config: A dictionary containing the model configuration.
            model: The model object.
            model_type: The type of the model object.
            input_shapes_and_types: A dictionary containing the input shapes and types of the model.
            output_shapes_and_types: A dictionary containing the output shapes and types of the model.
        """
        
        self.config = config
        self.model = model
        self.model_type = model_type
        self.input_shapes = input_shapes
        self.output_shapes = output_shapes
        self.parsed_input_shapes_and_types = {}
        self.parsed_output_shapes_and_types = {}
        
    @classmethod
    def init(cls, config: dict, model, model_type: str, input_shapes: dict, output_shapes: dict) -> 'BaseModel':
        """ A constructor-like function to init a base model object.
        
        Args:
            config: A dictionary containing the model configuration.
            model: The model object.
            model_type: The type of the model object.
            input_shapes: A dictionary containing the input shapes and types of the model.
            output_shapes: A dictionary containing the output shapes and types of the model.
        """
            
        return cls(config, model, model_type, input_shapes, output_shapes)
    
    def __type__(self):
        """ Return the type of the model object. """
        
        return str(self.model_type)
        
    def __str__(self):
        """ Print the model configuration. """
        
        return str(self.config)
    
    def __repr__(self):
        """ Print the model configuration. """
        
        return str(self.config)
    
    def __eq__(self, other):
        """ Check if two model objects are equal. """
        
        return self.config == other.config and self.model
    
    def convert_to_javascript(self):
        """ 
        Convert the model to a javascript format. 
        Dummy function for inheriting classes to implement.
        """

        pass
    
    def insert_to_webpage(self, web_template, converted_code):
        """ 
        Insert the model to a webpage. 
        We assume that the model is already converted to a javascript format, and its javascript code is in some sort of dictionary format.
        Dummy function for inheriting classes to implement.
        """
        
        pass
    


import tensorflow as tf
import numpy as np
import tf2onnx
import os, shutil

def randomly_construct_tf_model():
    """ 
    Randomly construct a tensorflow model object.
    This is done with keras.
    
    let's say inputs can be 2D, 3D or 4D tensors.
    if it is a 2D tensor, then it is a batch of 1D tensors. size: (batch_size, width)
    if it is a 3D tensor, then it is a batch of 2D tensors. size: (batch_size, width, height)
    if it is a 4D tensor, then it is a batch of 3D tensors. size: (batch_size, width, height, channels)
    let's assume batch size is always 1.
    """
    
    random_number_of_inputs = np.random.randint(1, 10)
    random_number_of_outputs = np.random.randint(1, 10)
    random_input_shapes = {}
    random_output_shapes = {}
    
    for i in range(random_number_of_inputs):
        random_number_of_dimensions = np.random.randint(2, 5)
        random_shape = [1]
        for j in range(random_number_of_dimensions):
            random_shape.append(np.random.randint(1, 10))
        random_shape = tuple(random_shape)
        random_input_shapes["input_" + str(i)] = random_shape
    
    for i in range(random_number_of_outputs):        
        random_number_of_dimensions = np.random.randint(2, 5)
        random_shape = [1]
        for j in range(random_number_of_dimensions):
            random_shape.append(np.random.randint(1, 10))
        random_shape = tuple(random_shape)
        random_output_shapes["output_" + str(i)] = random_shape
        
    # now we need to build a model that conforms to the input and output shapes
    # let's just do this for now:
    # model takes all input, flatten them, linear them all to a tensor with size [1, 42]
    # then this tensor is linear to all output.

    inputs = []
    for i in range(random_number_of_inputs):
        curr_input = tf.keras.Input(shape=random_input_shapes["input_" + str(i)])
        inputs.append(curr_input)
    flattened_inputs = []
    for i in range(random_number_of_inputs):
        curr_flattened_input = tf.keras.layers.Flatten()(inputs[i])
        flattened_inputs.append(curr_flattened_input)
    concatenated_inputs = tf.keras.layers.concatenate(flattened_inputs)
    linear_to_2 = tf.keras.layers.Dense(2)(concatenated_inputs)
    outputs = []
    for i in range(random_number_of_outputs):
        curr_output = tf.keras.layers.Dense(np.prod(random_output_shapes["output_" + str(i)]))(linear_to_2)
        curr_output = tf.keras.layers.Reshape(random_output_shapes["output_" + str(i)])(curr_output)
        outputs.append(curr_output)
    
    model = tf.keras.Model(inputs=inputs, outputs=outputs)

    return model, random_input_shapes, random_output_shapes

def randomly_construct_onnx_model():
    """
    Randomly construct an onnx model object.
    This function is converting a randomly constructed tensorflow model to onnx.
    """
    
    def get_dict(obj):
        if isinstance(obj, dict):
            return obj
        if isinstance(obj, tuple):
            for item in obj:
                if isinstance(item, dict):
                    return item
        if isinstance(obj, list):
            for item in obj:
                if isinstance(item, dict):
                    return item
    
    input_signature = []
    model, tmp1, tmp2 = randomly_construct_tf_model()
    
    for key in tmp1:
        value = tmp1.get(key)
        value = list(value)
        value.insert(0, None)
        value = tuple(value)
        input_signature.append(tf.TensorSpec(value, tf.float32, name=key))
    
    onnx_model, _ = tf2onnx.convert.from_keras(model, input_signature=input_signature, opset=13)
    return onnx_model, tmp1, tmp2

# model, _, _ = randomly_construct_tf_model()
model, _, _ = randomly_construct_onnx_model()


import tensorflow as tf
import numpy as np
import os
import shutil
from base_model import BaseModel

from test import randomly_construct_tf_model

tf_dtype_to_np_dtype = {
    tf.float16: np.float16,
    tf.float32: np.float32,
    tf.float64: np.float64,
    tf.int8: np.int8,
    tf.int16: np.int16,
    tf.int32: np.int32,
    tf.int64: np.int64,
    tf.uint8: np.uint8,
    tf.uint16: np.uint16,
    tf.uint32: np.uint32,
    tf.uint64: np.uint64,
}

class TensorflowKerasModel(BaseModel):
    """Tensorflow Keras Model object.

    Args:
        BaseModel (BaseModel): The base model object.
    """
    def __init__(self, config: dict, model, input_shapes: dict, output_shapes: dict):
        """ Initialize a tensorflow keras model object.

        Args:
            config: A dictionary containing the model configuration.
            model: The model object.
            input_shapes: A dictionary containing the input shapes and types of the model.
            output_shapes: A dictionary containing the output shapes and types of the model.
        """
        
        super().__init__(config, model, "Tensorflow-KerasModel", input_shapes, output_shapes)
    
    @classmethod
    def init(cls, config: dict, model) -> 'TensorflowKerasModel':
        """ A constructor-like function to init a tensorflow keras model object.

        Args:
            config: A dictionary containing the model configuration.
            model: The model object.
            input_shapes_and_types: A dictionary containing the input shapes and types of the model.
            output_shapes_and_types: A dictionary containing the output shapes and types of the model.
        """
        
        # convert to a tensorflow savedmodel
        if os.path.exists("tmp/temp_saved_model"):
            shutil.rmtree("tmp/temp_saved_model")
        else:
            os.makedirs("tmp/temp_saved_model")
            
        tf.saved_model.save(model, "tmp/temp_saved_model")
        model = tf.saved_model.load("tmp/temp_saved_model")
        
        # delete the temp saved model
        # TODO: refactor this to be safer
        shutil.rmtree("tmp/temp_saved_model")
        
        def get_dict(obj):
                if isinstance(obj, dict):
                    return obj
                if isinstance(obj, tuple):
                    for item in obj:
                        if isinstance(item, dict):
                            return item
                if isinstance(obj, list):
                    for item in obj:
                        if isinstance(item, dict):
                            return item
                        
        for signature in model.signatures.values():
            input_shapes = signature.structured_input_signature
            output_shapes = signature.structured_outputs
            break
                    
        input_shapes = get_dict(input_shapes)
        output_shapes = get_dict(output_shapes)
            
        input_specs = []
        output_specs = []
        
        for key in input_shapes:
            value = input_shapes.get(key)
            input_specs.append({
                'name': key,
                'shape': value.shape.as_list(),
                'dtype': tf_dtype_to_np_dtype[value.dtype]
            })
        
        for key in output_shapes:
            value = output_shapes.get(key)
            output_specs.append({
                'name': key,
                'shape': value.shape.as_list(),
                'dtype': tf_dtype_to_np_dtype[value.dtype]
            })
        
        return cls(config, model, input_specs, output_specs)
    
if __name__ == "__main__":
    model, input_shapes, output_shapes = randomly_construct_tf_model()
    model.save("test_model")
    model = tf.keras.models.load_model("test_model")
    kerasModel = TensorflowKerasModel.init({}, model)
    print(kerasModel.input_shapes)
    print(kerasModel.output_shapes)


# Class watchdog
class watchdog:
    # Initialize a watchdog instance
    def __init__(self, config: dict):
        self.config = config

    # A constructor-like function to init a watchdog instance
    @classmethod
    def init(cls, config: dict) -> 'watchdog':
        return cls(config)

# # Test block
# if __name__ == "__main__":
#     # Example dictionary as config
#     dictionary = {"key1": "value1", "key2": "value2"}
#
#     # Create a watchdog instance using the init method
#     model = watchdog.init(dictionary)
#
#     # Access and print the config attribute to verify it's working
#     print("model config:", model.config)



import tensorflow as tf
import numpy as np
import torch

def get_model_type(obj):
    """ Get the type of the model object.

    Args:
        obj: The model object.
        We assume the model object is either a tensorflow, tensorflow-lite or pytorch model.
    """
    
    if isinstance(obj, tf.keras.Model):
        return "tensorflow-keras"
    elif hasattr(obj, "signatures"):
        return "tensorflow"
    elif isinstance(obj, tf.lite.Interpreter):
        return "tensorflow-lite"
    elif isinstance(obj, torch.nn.Module):
        return "pytorch"
    else:
        raise TypeError("The model object is not a tensorflow, tensorflow-lite or pytorch model.")


import tensorflow as tf
from get_type import *
import numpy as np
import torch

def get_input_shape(obj):
    """ Get the model input tensor shape.

    Args:
        obj: The model object.
        We assume the model object is either a tensorflow, tensorflow-lite or pytorch model.
    """

    model_type = get_model_type(obj)

    if model_type == "tensorflow":
        return obj.get_layer(index = 0).input_shape
    elif model_type == "tensorflow-lite":
        return obj.get_input_details()[0]["shape"]
    elif model_type == "pytorch":
        pass


import numpy as np
import os
import shutil
from base_model import BaseModel
import onnxruntime as ort
import onnx
from test import randomly_construct_onnx_model

onnx_dtype_to_np_dtype = {
    'tensor(float16)': np.float16,
    'tensor(float)': np.float32,
    'tensor(double)': np.float64,
    'tensor(int8)': np.int8,
    'tensor(int16)': np.int16,
    'tensor(int32)': np.int32,
    'tensor(int64)': np.int64,
    'tensor(uint8)': np.uint8,
    'tensor(uint16)': np.uint16,
    'tensor(uint32)': np.uint32,
    'tensor(uint64)': np.uint64,
}

class OnnxModel(BaseModel):
    """Onnx Model object.

    Args:
        BaseModel (BaseModel): The base model object.
    """
    def __init__(self, config: dict, model, input_shapes: dict, output_shapes: dict):
        """ Initialize an onnx model object.

        Args:
            config: A dictionary containing the model configuration.
            model: The model object.
            input_shapes: A dictionary containing the input shapes and types of the model.
            output_shapes: A dictionary containing the output shapes and types of the model.
        """
        
        super().__init__(config, model, "OnnxModel", input_shapes, output_shapes)
    
    @classmethod
    def init(cls, config: dict, model) -> 'OnnxModel':
        """ A constructor-like function to init a onnx model object.

        Args:
            config: A dictionary containing the model configuration.
            model: The model object.
            input_shapes_and_types: A dictionary containing the input shapes and types of the model.
            output_shapes_and_types: A dictionary containing the output shapes and types of the model.
        """
        if os.path.exists("tmp/temp_saved_model"):
            shutil.rmtree("tmp/temp_saved_model")
            os.makedirs("tmp/temp_saved_model")
        else:
            os.makedirs("tmp/temp_saved_model")
        
        onnx.save(model, "tmp/temp_saved_model/model.onnx")
        sess = ort.InferenceSession("tmp/temp_saved_model/model.onnx")
        input_specs = []
        output_specs = []
        
        def format_shape(shape):
            for i in range(len(shape)):
                if isinstance(shape[i], str) and shape[i].startswith("unk__"):
                    shape[i] = None
            return shape
        
        for item in sess.get_inputs():
            input_specs.append({
                "name": item.name,
                "shape": format_shape(item.shape),
                "dtype": onnx_dtype_to_np_dtype[item.type]
            })
        for item in sess.get_outputs():
            output_specs.append({
                "name": item.name,
                "shape": format_shape(item.shape),
                "dtype": onnx_dtype_to_np_dtype[item.type]
            })
        
        return cls(config, model, input_specs, output_specs)
    
if __name__ == "__main__":
    model, input_shapes, output_shapes = randomly_construct_onnx_model()
    onnxModel = OnnxModel.init({}, model)
    print(onnxModel.input_shapes)
    print(onnxModel.output_shapes)